//------------------------------------------------------------------------------------
///	\file	mayaDoxygen.cpp
/// \date	8-Jul-2004
///	\author	Rob Bateman			robthebloke@hotmail.com
///	\brief	This plugin is designed to act in a similar way to doxygen but for maya 
///			scene files. The idea is that this script will generate a webpage for the
///			list of input scene files it gives it.
///
/// \note
/// 		This sourcefile is provided as-is and no guarentee or warrenty 
///			is provided for any damage that may arise from using it. This
///			code is free for your own use, the only thing i ask is a small
///			credit somewhere for my work. An e-mail saying you found it 
///			useful would also be much appreciated by myself ;)
///
///			If you find any bugs arising to maya version changes or you
///			have improvements to the source code, then feel free to mail
///			them to me and i'll update the source-code on the website.
///			Before you e-mail bug reports, make sure that it is my code 
///			causing the problem, and not some wierd oddity in the maya 
///			scene. This code is designed to handle most common model set-ups,
///			it is not designed to replicate perfectly everything that maya
///			does. 
/// 
///			DO NOT send me feature requests, i'm not likely to add to this
///			code myself unless something really new and groovy comes along.
///			This is mainly due to my current time restrictions. If you do
///			have code written that adds something fairly groovy to the 
///			exporter, then i will happily update the exporter code and give
///			full credit where due.....
/// 
///			Robert Bateman
/// 
///				email code to:    rbateman@bmth.ac.uk
///				contact:          robthebloke@hotmail.com
///
///
//------------------------------------------------------------------------------------

#ifdef WIN32
	#ifndef NT_PLUGIN
		#define NT_PLUGIN
	#endif

	// disable VC6 template name warning
	#pragma warning(disable:4786)

#endif


// link to some fairly funky maya libs. NOTE: The Image library only became part of maya
// from version 4.01 onwards. It is not available in version 4.0
//
#ifdef WIN32
#pragma comment(lib,"Foundation.lib")
#pragma comment(lib,"OpenMaya.lib")
#pragma comment(lib,"Image.lib")
#pragma comment(lib,"OpenMayaAnim.lib")
#endif

// include each header for each class you need. Gets a bit laborious later on...
//
#include <maya/MFloatMatrix.h>
#include <maya/MMatrix.h>
#include <maya/MLibrary.h>
#include <maya/MPlug.h>
#include <maya/MFnDagNode.h>
#include <maya/MPointArray.h>
#include <maya/MFloatVectorArray.h>
#include <maya/MFnMesh.h>
#include <maya/MFnNurbsSurface.h>
#include <maya/MFnTransform.h>
#include <maya/MFileIO.h>
#include <maya/MItDependencyNodes.h>
#include <maya/MItMeshPolygon.h>
#include <maya/MItMeshVertex.h>
#include <maya/MSimple.h>


#include <vector>
#include <map>
#include <set>
#include <string>
#include <fstream>
#include <sstream>
#include <iostream>


//------------------------------------------------------------------------------------
/// the type of comment block that can appear
///
#define DOX_AUTHOR 0
#define DOX_DATE 1 
#define DOX_TODO 2
#define DOX_BRIEF 3
#define DOX_NOTES 4
#define DOX_REQUIRES 5
#define DOX_LAST 6

typedef unsigned int DoxNoteType;

/// the Path to the output HTML folder
std::string HTML_PATH ;

//------------------------------------------------------------------------------------
/// the text string versions of the comment tags
const char *DoxNoteFlags[] = {
	"\\author",
	"\\date",
	"\\todo",
	"\\brief",
	"\\notes",
	"\\requires",
	NULL	
};

//------------------------------------------------------------------------------------
/// \brief	A nasty utility function to query if the specified string is a dox tag
/// \param	s	-	the string to test
/// \return	the string as a doxy tag
///
DoxNoteType GetDoxTag(std::string s)
{
	DoxNoteType dt = DOX_AUTHOR;
	
	for( ; dt != DOX_LAST; ++dt )
	{
		if( s == DoxNoteFlags[dt] )
			break;
	}
	
	return dt;
}

//------------------------------------------------------------------------------------
/// \brief	This structure holds a breakdown of the data contained within the notes
/// 		attribute of a node.
/// 
struct DoxCommentBlock {
	
	/// ctor
	DoxCommentBlock() {}

	/// dtor
	~DoxCommentBlock() {}

	/// copy ctor
	DoxCommentBlock(const DoxCommentBlock& dcb) 
		: Obj(dcb.Obj), Node(dcb.Node), TypeStr(dcb.TypeStr),
		  Author(dcb.Author),Date(dcb.Date),
		  Todo(dcb.Todo), Brief(dcb.Brief), Notes(dcb.Notes),
		  Requires(dcb.Requires) 
	{
	}
	
	/// The maya object this node refers to.
	MObject Obj;
	
	/// The name of the node that this comment block is attached to
	std::string Node;
	
	/// The typename of the node that this comment block belongs to.
	std::string TypeStr;
	
	/// the string following the \author tag
	std::string Author;
	
	/// the string following the \date tag
	std::string Date;
	
	/// the string following the \todo tag
	std::string Todo;
	
	/// the string following the \brief tag, or string if no tag is specified
	std::string Brief;
	
	/// the string following the more verbose \notes flag
	std::string Notes;
	
	/// the string following the \requires flag 
	std::string Requires;

	//--------------------------------------------------------------------------------
	/// \brief	The less than operator is overloaded in order to provide alphabetical
	/// 		sorting of the comments based on their node names.
	/// \param	dcb	-	the comment block to compare with
	/// \return true if less
	/// 
	bool operator < (const DoxCommentBlock& dcb) const {
		return Node < dcb.Node;
	}
	
	void WriteLink(std::ofstream& ofs,const std::string& file) const;
	
	/// \brief	This function will basically write out the main html file for this node. 
	/// 		
	void WriteMainHtml(const std::string& path,const std::string& file) const ;
};

//------------------------------------------------------------------------------------
/// \brief	Writes a link for a single node page
/// \param	ofs		-	the output file stream
/// \param	file	-	the output filename
///
void DoxCommentBlock::WriteLink(std::ofstream& ofs,const std::string& file) const {
	
	ofs << "<h3><a href=" 
		<< file << "_" 
		<< Node << ".html>"
		<< Node << "</a>""</h3></br><i>" 
		<< Brief << "</i></br></br>\n";
	
}

//------------------------------------------------------------------------------------
/// \brief	This function will basically write out the main html file for this node. 
/// \param	html	-	the html file name to write
/// \param	file	-	the maya binary file
/// 		
void DoxCommentBlock::WriteMainHtml(const std::string& html,const std::string& file) const {

	// create the html filename
	std::string Path ;
	Path = html;
	Path += "_";
	Path += Node;
	Path += ".html";
	std::ofstream ofs(Path.c_str());
	
	
	// if file opened
	if(ofs)
	{
		// write html header
		ofs << "<html><head><title>" 
			<< Node 
			<< "</title></head>\n"
			<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
			<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a>"
			<< "    <a HREF=\""
			<< html 
			<< "\">"
			<< file
			<< "</a>\n</PRE>\n";
		

		ofs << "<h1><center><font COLOR=\"#880011\">" << Node << "</font></center></h1>\n";
		
		ofs << "<h3>Type</h3><br>" << TypeStr << "<br><br>\n";
		
		// brief description
		if( Brief.size() > 0 )
			ofs << "<h3>Brief</h3><br>" << Brief << "<br><br>\n";
		
		// author name
		if( Author.size() > 0 )
			ofs << "<h3>Author</h3><br>" << Author << "<br><br>\n";
	
		// output date
		if( Date.size() > 0 )
			ofs << "<h3>Date</h3><br>" << Date << "<br><br>\n";
	
		// write the requirements
		if( Requires.size() > 0 )
			ofs << "<h3>Requires</h3><br>" << Requires << "<br><br>\n";
	
		// output the notes
		if( Notes.size() > 0 )
			ofs << "<h3>Notes</h3><br>" << Notes << "<br><br>\n";
	
		// output any todo stuff
		if( Todo.size() > 0 )
			ofs << "<h3>ToDo</h3><br>" << Todo << "<br><br>\n";

		// write some links at the bottom
		ofs	<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
			<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a>"
			<< "    <a HREF=\"" << html << "\">" << file << "</a>\n</PRE>\n";

		// end page
		ofs << "</body></html>\n";
		ofs.close();		
	}
}


//------------------------------------------------------------------------------------
/// \brief	This structure holds a breakdown of the data contained within the notes
/// 		attribute of a node.
/// 
struct DoxFile {
	
	/// 
	DoxFile() 
		: m_iMeshCount(0),
		  m_iNurbsCount(0),
		  m_iLightCount(0),
		  m_iMaterialCount(0)
	{
	}

	
	/// a set of the commented node
	std::set< DoxCommentBlock > m_NodeComments;
	
	/// \brief	This function 
	void BuildDoxCommentBlock( const MObject &obj );
	
	/// The name of the maya file
	std::string FileName;
	/// The name of the maya file
	std::string HTMLName;
	
	/// The directory path in which the file is contained
	std::string path;
	
	/// an array of the texture files used by this scene
	std::map< std::string, std::string > m_aTextureFiles;
	
	/// an array of the maya files referenced by this scene
	std::set< std::string > m_aReferenceFiles;
	
	/// General Info on the scene, ie num Meshes, Nurbs surfaces, materials etc.
	unsigned int m_iMeshCount;
	unsigned int m_iNurbsCount;
	unsigned int m_iMaterialCount;
	unsigned int m_iLightCount;
	
	void WriteMainHtml(const std::string& path);

	/// \brief	This function parses all the nodes in the scene looking for items of interest
	void ParseScene();
};

//------------------------------------------------------------------------------------
/// \brief	This function outputs the html file for the maya scene page
///	\param	path	-	the path to the html output directory.
/// 
void DoxFile::WriteMainHtml(const std::string& path) 
{
	// remove the directory before the html page
	std::string f;
	int i=FileName.size();
	while (--i>=0) {
		if (FileName[i] == '\\' || 
			FileName[i] == '/' ) {
			f = &(FileName[i+1]);
			break;
		}
	}

	// construct a html file path
	HTMLName = path;
	HTMLName += f;
	HTMLName += ".html";
	
	// open an output file stream (the output html)
	std::ofstream ofs(HTMLName.c_str());
	if( ofs )
	{
		// write the 
		ofs << "<html><head><title>"	// title
			<< FileName
			<< "</title></head>\n\n"
			<< "<PRE>\n"				// links to other main pages
			<< "<a HREF=\"index.html\">Index</a>    <a HREF=\"files.html\">Files</a>"
			<< "    <a HREF=\"todo.html\">To Do</a>\n</PRE>\n\n"
			<< "<h1><center><font COLOR=\"#880011\">"
			<< FileName 
			<< "</font></center></h1>\n<body>\n\n";
			
		// write a basic block of code that gives quick overview of scene data
		ofs << "<h2>Scene Info</h2>\n<PRE>"
			<< "\n    Number Of Meshes         = " << m_iMeshCount
			<< "\n    Number Of Nurbs Surfaces = " << m_iNurbsCount
			<< "\n    Number Of Materials      = " << m_iMaterialCount
			<< "\n    Number Of Lights         = " << m_iLightCount
			<< "\n</PRE>\n";
		
		// if texture files are present, add 
		if( m_aTextureFiles.size() > 0 )
		{
			std::map< std::string, std::string >::iterator ittex = m_aTextureFiles.begin();
			ofs << "<h2>Texture Files</h2>\n<PRE>\n"; 
			for( ; ittex != m_aTextureFiles.end(); ++ittex )
			{
				ofs << "    <b>" 
					<< ittex->first 
					<< ":</b>   <i>"
					<< ittex->second 
					<< "</i>"<<std::endl;
			}
			ofs << "</PRE>\n";
		}
		
		// write any referenced files
		if( m_aReferenceFiles.size() > 0 )
		{
			std::set< std::string >::iterator ittex = m_aReferenceFiles.begin();
			ofs << "<h2>Referenced Files</h2>\n<PRE>\n"; 
			for( ; ittex != m_aReferenceFiles.end(); ++ittex )
			{
				ofs << "    " << *ittex << std::endl;
			}
			ofs << "</PRE>\n";
		}
		
		// write the node comments
		if( m_NodeComments.size() > 0 )
		{
			ofs << "<h2>Scene Nodes</h2></br>\n";
			std::set< DoxCommentBlock >::iterator it = m_NodeComments.begin();
			for( ; it != m_NodeComments.end(); ++it )
			{
				it->WriteMainHtml(HTMLName,FileName);
				it->WriteLink(ofs,HTMLName);
			}
		}
		
		// write some links at the bottom.
		ofs	<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
			<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a>"
			<< "    <a HREF=\"" << FileName << "\">" << FileName << "</a>\n</PRE>\n";
	
		
		ofs << "</body></html>\n";
		ofs.close();		
	}
}


//------------------------------------------------------------------------------------
/// \brief	This function queries the maya node to see if there is a "notes" attribute
///			on it. If there is then it splits up the string into meaningful chunks.
/// \param	obj	-	the maya object
/// 
void DoxFile::BuildDoxCommentBlock( const MObject &obj )
{
	// attach a function set to the maya object
	MFnDependencyNode fn(obj);
	
	// status code
	MStatus stat;
	
	// try to see if the node has an attribute called "notes". This is what maya 
	// creates when you set add notes at the bottom of the attribute editor.
	MPlug plg = fn.findPlug("notes",&stat);
	
	// only do anything if we found the attribute
	if( stat == MS::kSuccess )
	{
		// create new comment block for node
		DoxCommentBlock* dcb = new DoxCommentBlock;
				
		// store the maya object 
		dcb->Obj = obj;
		
		// store name of node
		dcb->Node    = fn.name().asChar();
		
		// store type name of node
		dcb->TypeStr = fn.typeName().asChar();
		
		// next thing to do is to get the actual text value of the string. 
		MString text;
		plg.getValue(text);
		
		{
			std::string stext = text.asChar();
			std::string::const_iterator it = stext.begin();
			std::string* ptr_cur = &(dcb->Brief);
			std::string buffer;
			
			// need to parse the text to find the dox tags
			while( it < stext.end() )
			{
				if(	*it == '\0' )
					break;
				// skip white
				if( *it == ' '  ||
					*it == '\t' ||
					*it == '\n'  )
				{
					++it;
				}
				else
				{
					// read next string from comment
					while( it < stext.end() && 
						   *it != '\n' && 
					   	   *it != ' ' &&
						   *it != '\t' )
					{
						buffer += *it;
						++it;
					}
					++it;
					
					// check to see if the string is a dox tag
					DoxNoteType dt = GetDoxTag(buffer);

					// either re-assign our current pointer, or append the text
					// string to the current string being filled.
					switch( dt )
					{
					case DOX_AUTHOR:
						ptr_cur = &(dcb->Author);
						break;

					case DOX_DATE:
						ptr_cur = &(dcb->Date);
						break;

					case DOX_TODO:
						ptr_cur = &(dcb->Todo);
						break;

					case DOX_BRIEF:
						ptr_cur = &(dcb->Brief);
						break;

					case DOX_NOTES:
						ptr_cur = &(dcb->Notes);
						break;

					case DOX_REQUIRES:
						ptr_cur = &(dcb->Requires);
						break;

					default:
						*ptr_cur += buffer;
						*ptr_cur += " ";
						break;
					}
					buffer.resize(0);
				}
			}
		}
		
		// add it to the set
		m_NodeComments.insert( *dcb );
		
		// delete the unused data
		delete dcb;
	}
}

//----------------------------------------------------------------------------------------
/// \brief	This function parses all the nodes in the scene looking for items of interest
///
void DoxFile::ParseScene()
{
	// this bit build a list of all files referenced by this file.
	{
		MStringArray refs;
		MFileIO::getReferences( refs );
		
		for( int j=0;j<refs.length();++j)
		{
			m_aReferenceFiles.insert( refs[j].asChar() );
		}
	}
	
	// iterate through the scene nodes
	MItDependencyNodes itDep( MFn::kInvalid );
	for( ; !itDep.isDone(); itDep.next() )
	{
		// build a comment block for a laugh
		BuildDoxCommentBlock( itDep.item() );
		
		// see what type of node this is.....
		switch( itDep.item().apiType() )
		{
		// if a mesh is found in file
		case MFn::kMesh:
			{
				// see if the mesh is actually important and increment count if it is
				MFnMesh fnDep( itDep.item() );
				if( !fnDep.isIntermediateObject() )
					m_iMeshCount++;
			}
			break;
		
		// if we have a file texture node
		case MFn::kFileTexture:
			{
				MFnDependencyNode fn( itDep.item() );
				
				// get the file-texture-name attribute
				MPlug plug = fn.findPlug("ftn");
				MString str;
				
				// get the value of the attribute
				plug.getValue(str);
				
				// insert pair into map
				m_aTextureFiles.insert( 
					std::make_pair( std::string( fn.name().asChar() ), 
									std::string( str.asChar() ) ) );
			}
			break;
		
		// if we have a nurbs surface
		case MFn::kNurbsSurface:
			{
				MFnNurbsSurface fnDep( itDep.item() );
				if( !fnDep.isIntermediateObject() )
					m_iNurbsCount++;
			}
			break;

		// if we have found a light
		case MFn::kLight:
		case MFn::kPointLight:
		case MFn::kAmbientLight:
		case MFn::kDirectionalLight:
		case MFn::kSpotLight:
		case MFn::kAreaLight:
			{
				m_iLightCount++;
			}
			break;

		// if found a material...
		case MFn::kPhong:
		case MFn::kLambert:
		case MFn::kBlinn:
		case MFn::kPhongExplorer:
			{
				m_iMaterialCount++;
			}
			break;
			
		default: 
			break;
		}
	}
	
	// write the html file for this maya scene file.
	WriteMainHtml(HTML_PATH);	
}

// build up an info list for each file in the directory
std::vector< DoxFile* > FileList;

void Clean()
{
	std::vector< DoxFile* >::iterator it = FileList.begin();
	for(;it!=FileList.end();++it)
	{
		delete *it;
	}
	FileList.clear();
}


//--------------------------------------------------------------------------
/// \brief	This function outputs a todo list
///
void DoxOutputToDoPage()
{
	// make a filename
	std::string file = HTML_PATH;
	file += "todo.html";
	
	// create an output file
	std::ofstream ofs( file.c_str() );

	// write html header
	ofs << "<html><head><title>To Do List</title></head>\n"
		<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
		<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a></PRE>\n";

	// write todo list title
	ofs << "<h1><center><font COLOR=\"#880011\">To Do List</font></center></h1>\n";

	// write anything todo in our files
	std::vector< DoxFile* >::iterator itFile = FileList.begin();
	for( ; itFile != FileList.end(); ++itFile )
	{
		ofs << "<h2><a href=\"" << (*itFile)->HTMLName 
			<< "\">" << (*itFile)->FileName << "</a></h2><PRE>\n";
		
		// get the comments
		std::set< DoxCommentBlock >::iterator itc = (*itFile)->m_NodeComments.begin();
		
		for( ; itc != (*itFile)->m_NodeComments.end(); ++itc )
		{
			if( itc->Todo.size() > 0 )
			{
				ofs << "    <b><a href=\""  << (*itFile)->HTMLName << "_" << itc->Node 
					<< ".html\">" << itc->Node  << "</a>"
					<< "</b><br>        <i>" << itc->Todo << "\n";
			}
		}
		
		ofs << "</PRE>\n";
	}

	ofs	<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
		<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a></PRE>\n";	
	
	ofs.close();
}

//--------------------------------------------------------------------------
/// \brief	This function simply writes a html page detailing all of the maya
///			files.
///
void DoxOutputFilePage()
{
	// create a filename
	std::string file = HTML_PATH;
	file += "files.html";
	
	// open a file
	std::ofstream ofs( file.c_str() );
	
	// write html header
	ofs << "<html><head><title>FileList</title></head>\n"
		<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
		<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a></PRE>\n";

	// write title
	ofs << "<h1><center><font COLOR=\"#880011\">Scene File List</font></center></h1>\n";

	// write each file as a link
	std::vector< DoxFile* >::iterator itFile = FileList.begin();
	for( ; itFile != FileList.end(); ++itFile )
	{
		ofs << "<h2><center><a href=\""  << (*itFile)->HTMLName 
			<< "\">" << (*itFile)->FileName  << "</a>"
			<< "</h2></center><br><br>\n";
	}
	
	// write some links at the bottom
	ofs	<< "<PRE>\n<a HREF=\"index.html\">Index</a>    "
		<< "<a HREF=\"todo.html\">To Do</a>    <a HREF=\"files.html\">Files</a></PRE>\n";
	
	// close the file
	ofs.close();
}

//--------------------------------------------------------------------------
/// \brief	This function simply writes a html index page
///
void DoxOutputIndexPage()
{
	std::string file = HTML_PATH;
	file += "index.html";
	
	std::ofstream ofs( file.c_str() );
	
	ofs << "<html><head><title>Index</title></head>\n";

	ofs << "<h1><center><font COLOR=\"#880011\">Index</font></center></h1><br><br>\n";
	ofs << "<h2><center><a href=\"todo.html\">To Do List</a></center></h2><br><br>\n";
	ofs << "<h2><center><a href=\"files.html\">Scene File List</a></center></h2><br><br>\n";

	ofs.close();
}

//------------------------------------------------------------------------------------
DeclareSimpleCommand(doxymaya,"Rob Bateman","1.0");
MStatus doxymaya::doIt(const MArgList& arg)
{
	// holds a list of all of the input files
	std::vector< std::string > Files;
	
	// loop through all of the arguments
	for( unsigned int i = 0; i < arg.length(); ++i )
	{
		// get the arg
		std::string temp = arg.asString(i).asChar();
		
		// if input file
		if( temp == "-f" )
		{
			// check for error
			if( ++i == arg.length() ) {
				displayError("ERROR specifying arguments\n");
				return MS::kFailure;
			}
			
			// get arg value
			temp = arg.asString(i).asChar();
			
			// store filepath
			Files.push_back( temp );
			
		}
		else
		if( temp == "-od" )
		{
			// check for error
			if( ++i == arg.length() ) {
				displayError("ERROR specifying arguments\n");
				return  MS::kFailure;
			}

			// get path
			HTML_PATH = arg.asString(i).asChar();	

			// make sure the path is terminated properly
			if( HTML_PATH[HTML_PATH.size()-1] != '/' )
			{
				HTML_PATH += "/";
			}						
		}
		
	}
	
	// iterate through all of the input files
	std::vector< std::string >::iterator it =  Files.begin();
	for( ; it != Files.end(); ++it ) {

		// get filename
		std::string s = *it;

		// if maya binary of maya ascii
		if( s.size() && 
			(s[ s.size()-1 ] == 'b' || s[ s.size()-1 ] == 'a') && 
			s[ s.size()-2 ] == 'm' && 
			s[ s.size()-3 ] == '.' )
		{
			// create a new dox file reference
			DoxFile* pd = new DoxFile;

			// add it to the list with the others
			FileList.push_back( pd );

			// open the maya file
			MFileIO::open( s.c_str() );

			// store filename and path
			pd->FileName = *it;

			// parse this scene and write out the html file.
			pd->ParseScene();
		}
	}

	// write the main index page
	DoxOutputIndexPage();
	// write the file page
	DoxOutputFilePage();
	// write the todo page
	DoxOutputToDoPage();

	Clean();


	return MS::kSuccess;
}

